/*
    Fichier gentexte.c
    Auteur Bernard Chardonneau

    Logiciel libre, droits d'utilisation prciss en franais
    dans le fichier : licence-fr.txt

    Traductions des droits d'utilisation dans les fichiers :
    licence-en.txt , licence-es.txt , licence-pt.txt ,
    licence-eo.txt , licence-eo-utf.txt

    Droits d'utilisation galement sur la page web :
    http://cyloop.tuxfamily.org/voir.php?page=droits


    Bibliothque de fonctions destines  rajouter du texte
    dans les images gnres par cylgraph.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "types.h"
#include "messages.h"
#include "image.h"
#include "gentexte.h"


// variables globales  cette bibliothque

// tables UTF-8 en mmoire

// une table par valeur du 1er caractre entre C0 et DF
static octet *table_utf [32];

// un fichier par valeur du 1er caractre entre E0 et FF
static FILE *fic_utf [32];

// indique quelles valeurs des 2 tables prcdentes ont t initialises
static int  init_utf [64] =
            { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
              0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

// autres variables globales

// descripteur de fichier pour les fontes charges en mmoire
FILE   *ficfonte;

// mmorise si on a charg la bonne fonte iso-8858-n
octet  fonte_base_OK = 0;

// jeu de caractres ASCII tendu (caractres sur un octet)
// les caractres de controle ne seront pas affichs
extern octet fonte [224][hauteur_fonte];



/* indique si le caractre pass en paramtre est un caractre hexadcimal */

int carhexa (char caract)
{
    if ('0' <= caract && caract <= '9')
        return 1;

    caract = caract | 0x20;

    return ('a' <= caract && caract <= 'f');
}



/* retourne la valeur du caractre hexadcimal pass en paramtre */

int valhexa (char caract)
{
    if ('0' <= caract && caract <= '9')
        return (caract & 0x0F);
    else
        return (caract & 0x0F) + 9;
}



/*
   saute les caractres blancs ou tabulations
   ventuels  partir de la position courante

   retourne :
   - la valeur du code hexadcimal qui suit, si c'en est un
   - -1 si les caractres qui suivent n'expriment pas un code hexadcimal
   - -2 si fin de fichier
*/

int lecvalhexa ()
{
    int caract;  // caractre du fichier fonte
    int valeur;  // valeur hexadcimale correspondante


    // lire un caractre
    caract = getc (ficfonte);

    // termin si fin de fichier
    if (caract == EOF)
        return -2;

    // tant qu'on est sur un blanc ou une tabulation
    while (caract == ' ' || caract == '\t')
        // lire le caractre suivant
        caract = getc (ficfonte);

    // si on est tomb sur un caractre hexadcimal
    if (carhexa (caract))
    {
        // rcuprer la valeur correspondant 
        // 2 caractres hexadcimaux successifs
        valeur = valhexa (caract);

        caract = getc (ficfonte);

        if (carhexa (caract))
            valeur = (valeur << 4) | valhexa (caract);
        else
            // -1 si on n'avait pas 2 caractres hexadcimaux successifs
            valeur = -1;
    }
    // sinon
    else
    {
        // on n'a pas trouv de valeur hexadcimale
        valeur = -1;

        // sauter le reste de la ligne
        while (caract != '\n' && caract != EOF)
            caract = getc (ficfonte);
    }

    // retourner la valeur rcupre ou le code d'erreur
    return (valeur);
}



/* retourne le nom du rpertoire contenant les fichiers fonte */

char *dirfonte ()
{
    static char *dirf = NULL;  // static pour ne faire la recherche qu'une fois


    // si ce rpertoire n'a pas encore t cherch
    if (! dirf)
    {
        // si les commandes de cyloop sont implantes dans /usr/bin
        if (strcmp (dircom (), "/usr/bin") == 0)
        {
            // les fontes sont dans /usr/share/cyloop/fontes
            dirf = malloc (25);
            strcpy (dirf, "/usr/share/cyloop/fontes");
        }
        // sinon, les fontes sont dans les rpertoire fontes situ au
        // mme niveau que le rpertoire bin des commandes excutables
        else
        {
            // calcul de la taille de mmoir ncessaire et allocation
            dirf = malloc (strlen (dircom ()) + 3);

            // copie du rpertoire d'implantation des commandes de cyloop
            strcpy (dirf, dircom ());

            // on remplace bin par fontes dans la chaine prcdente
            strcpy (dirf + strlen (dirf) - 3, "fontes");
        }
    }

    // retourner le rsultat
    return (dirf);
}



/*
   extrait dans le fichier fonte UTF-8 identifi par carbase
   la reprsentation graphique de caractre correspondant 
   la squence suiteseq et mmorise le rsulat dans dessincar
*/

void recupfonte (octet carbase, octet *suiteseq, octet *dessincar)
{
    int   pospremier; // position du premier caractre dans suiteseq
    octet car2;       // 2me caractre de la squence UTF-8
    octet car3;       // 3me caractre de la squence UTF-8
    int   numcar;     // compteur
    int   val;        // valeur code en hexadcimal dans le fichier
    int   i, j;       // compteurs



    printf ("Appel de recupfonte (%02X ", carbase);

    // position dans le tableau suiteseq du 2me caractre de la squence
    // (l'ordre des caractres est invers dans suiteseq)
    if (carbase < 0xF0)
        pospremier = 1;
    else
        pospremier = 2;

    if (pospremier == 1)
        printf ("%02X %02X)\n", suiteseq [1], suiteseq [0]);
    else
        printf ("%02X %02X %02X)\n", suiteseq [2], suiteseq [1], suiteseq [0]);

    // initialisation dessincar  blanc
    for (i = 0; i < hauteur_fonte; i++)
        dessincar [i] = 0;

    // slectionner le fichier fonte  lire
    ficfonte = fic_utf [carbase & 0x1F];

    // rembobinage du fichier
    rewind (ficfonte);

    // pour viter d'accder souvent  un tableau
    car2 = suiteseq [pospremier];
    car3 = suiteseq [pospremier - 1];

    // recherche dans le fichier fonte
    do
    {
        // lire la premire valeur hexadcimale de la ligne
        val = lecvalhexa ();

        // si les 2 premires valeurs hexadcimales
        // de la ligne sont celles recherches
        if ((val == car2) && (lecvalhexa () == car3))
        {
            // ainsi que le 3me si elle existe
            if ((carbase < 0xF0) || (lecvalhexa () == *suiteseq))
            {
                // rcuprer l'image graphique du caractre
                for (i = 0; i < hauteur_fonte; i++)
                    dessincar [i] = lecvalhexa ();

                // recherche du caractre termine avec succs
                return;
            }
        }

        // passage  la ligne suivante du fichier fonte

        // rpter
        do
        {
            // passer au caractre suivant
            val = getc (ficfonte);

            // si fin de fichier, sortir
            if (val == EOF)
                return;
        }
        // jusque changement de ligne
        while (val != '\n');
    }
    while (i != -2);  // on arrte en fin de fichier fonte
}



/*
   charge en mmoire un fichier fonte

   paramtres : - adresse de la table contenant le dessin des caractres
                - valeur du premier caractre mmoris dans la table

   le fichier fonte a t ouvert par chargefonte_iso ou chargefonte_utf
   on utilise le descripteur de fichier obtenu lors de cette ouverture
*/

void chargefonte (octet fonte [][hauteur_fonte], octet valdeb)
{
    int  val;   // valeur code en hexadcimal dans le fichier
    int  i, j;  // compteurs


    printf ("Appel de chargefonte (%08X, %02X)\n", fonte, valdeb);

    // sauter les lignes de commentaires dans le fichier fonte
    do
        i = lecvalhexa ();
    while (i == -1);

    // tant qu'on trouve des lignes commenant par un code hexadcimal
    while (i >= 0)
    {
        // si le code correspond  celui d'un caractre autoris
        if (i >= valdeb)
        {
            // calcul de la position de ce code dans la table
            i = i - valdeb;

            // dcoder et recopier dans la table les valeurs
            // correspondant au dessin du caractre
            j = 0;

            do
            {
                val = lecvalhexa ();

                if (val >= 0 && j < hauteur_fonte)
                    fonte [i][j++] = val;
            }
            while (val >= 0);
        }
        // sinon message d'erreur
        else
            // "Prsence du caractre %02X non autoris dans le fichier fonte"
            aff_err_argnum ("CARFONTE_INTERDIT", i);

        // rechercher une autre ligne avec un code de caractre  traiter
        do
            i = lecvalhexa ();
        while (i == -1);
    }
}



/*
   ouvre un fichier fonte UTF-8 et le charge en mmoire
   s'il concerne des squences UTF-8 de 2 caractres

   pour les squences UTF-8 sur 3 ou 4 caractres
   on garde en mmoire le descripteur du fichier

   paramtre : premier caractre de la squence UTF-8
*/

void chargefonte_utf (octet carbase)
{
    int   postable;    // numro de l'lment de table_utf charg en mmoire
    char  *chemfonte;  // chemin d'accs au fichier fonte
    octet *sous_table; // image en mmoire du fichier charg


    printf ("Appel de chargefonte_utf (%02X)\n", carbase);

    // fabriquer le chemin d'accs au fichier fonte
    chemfonte = malloc (strlen (dirfonte ()) + 9);
    sprintf (chemfonte, "%s/utf8-%02x", dirfonte (), carbase);

    // position dans la table des fontes UTF-8 charges en mmoire
    // ou dans la liste des descripteurs de fichiers
    postable = carbase & 0x1F;

    // si les squences UTF-8 concernes par le
    // fichier fonte tiennent sur 2 caractres
    if (carbase < 0xE0)
    {
        // cas gnral
        if (carbase > 0xC3 || carbase == 0xC0)
        {
            // allouer une zone mmoire pour stocker la sous
            // table UTF-8 mmorise dans le fichier fonte
            sous_table = malloc (64 * hauteur_fonte);

            // mmoriser l'adresse mmoire dans table_utf
            table_utf [postable] = sous_table;

            // si allocation mmoire russie
            if (sous_table)
            {
                // initialiser cette zone mmoire
                memset (sous_table, 0, 64 * hauteur_fonte);

                // cas trs particulier :
                // le premier caractre de la squence UTF-8 est C0
                // (ne devrait jamais se produire comme pour le caractre C1
                //  car les caractres concerns peuvent tre cods sur un
                //  octet, mais il n'est pas difficile de traiter ce cas)
                if (carbase == 0xC0)
                    // recopier le morceau de la table ASCII
                    // (ou ISO-8859-1) concernant ces caractres
                    memcpy (sous_table + (32 * hauteur_fonte), fonte,
                                                  32 * hauteur_fonte);
            }
            // sinon message d'erreur
            else
            {
                // "Manque de place mmoire"
                // "certains caractres seront remplacs par des blancs\n"
                affiche_err ("MANQUE_MEMOIRE");
                affiche_err ("CAR_RPL_BLANCS");

                // plus la peine de lire le fichier fonte dans ce cas
                return;
            }
        }
        // sinon la sous table UTF-8 correspond  une
        // partie de la table ISO-8859-1 dj en mmoire
        else
            table_utf [postable] = fonte [(64 * postable) - 32];

        // ouvrir le fichier fonte en lecture
        ficfonte = fopen (chemfonte, "r");

        // si ce fichier a pu tre ouvert
        if (ficfonte)
            // charger son contenu en mmoire
            chargefonte ((void *) table_utf [postable], 0x80);

        // sinon message d'erreur si la fonte  rcuprer
        // n'est pas une partie de la table ISO-8859-1
        else if (carbase > 0xC3)
        {
            // "Fichier fonte %s manquant\n"
            // "certains caractres seront remplacs par des blancs\n"
            aff_err_arg ("FICFONTE_MANQUANT", chemfonte);
            affiche_err ("CAR_RPL_BLANCS");
        }
    }
    // sinon, les squences UTF-8 concernes par le
    // fichier fonte tiennent plus de 2 caractres
    else
    {
        // ouvrir le fichier fonte en lecture et mmoriser son descripteur
        fic_utf [postable] = fopen (chemfonte, "r");

        // message d'erreur si ce fichier n'a pas pu tre ouvert
        if (!fic_utf [postable])
        {
            // "Fichier fonte %s manquant\n"
            // "certains caractres seront remplacs par des blancs\n"
            aff_err_arg ("FICFONTE_MANQUANT", chemfonte);
            affiche_err ("CAR_RPL_BLANCS");
        }
    }
}



/*
   charge un fichier fonte pour remplacer
   certains caractres du jeu iso-8859-1

   paramtre : nom du fichier fonte  charger en mmoire
*/

void chargefonte_iso (char *nomfonte)
{
    int  numiso;     // 2me numro de la fonte ISO slectione, -1 si non ISO
    int  debinit;    // numro de 1er caractre du tableau fonte  initialiser
    char *chemfonte; // chemin d'accs au fichier fonte
    int  i, j;       // compteurs


    printf ("Appel de chargefonte_iso (%s)\n", nomfonte);

    // si fonte iso-8859
    if (memcmp (nomfonte, "iso8859", 7) == 0)
        // extraire le numro
        numiso = atoi (nomfonte + 8);
    // sinon
    else
        // autre cas
        numiso = -1;

    // dterminer  partir de quelle case de la table
    // iso8859-1 on va rinitialiser les valeurs
    switch (numiso)
    {
        // rinitalisation  partir de la valeur A0h
        case  5 :
        case  6 :
        case 11 : debinit = 128;
                  break;

        // rinitalisation  partir de la valeur C0h
        case  7 :
        case  8 : debinit = 160;
                  break;

        // pas de rinitialisation : crasement lorsque caractre diffrent
        default : debinit = 224;
                  break;
    }

    // rinitialiser tout ou partie de la table selon la valeur de debinit
    for (i = debinit; i < 224; i++)
        for (j = 0; j < hauteur_fonte; j++)
            fonte [i][j] = 0;

    // fabriquer le chemin d'accs au fichier fonte
    chemfonte = malloc (strlen (dirfonte ()) + strlen (nomfonte) + 2);
    sprintf (chemfonte, "%s/%s", dirfonte (), nomfonte);

    // ouvrir ce fichier en lecture
    ficfonte = fopen (chemfonte, "r");

    // message d'erreur si fichier non trouv
    if (! ficfonte)
    {
        // sauf si on utilise le jeu ISO-8859-1
        if (numiso != 1)
        {
            // "Fichier fonte %s manquant\n"
            aff_err_arg ("FICFONTE_MANQUANT", chemfonte);

            // on dite le message d'erreur le mieux adapt

            if (numiso == -1 || debinit == 96)
                // "certains caractres seront remplacs par des blancs\n"
                affiche_err ("CAR_RPL_BLANCS");

            else if (debinit == 224)
            // "certains caractres seront remplacs par ceux du jeu ISO-8859-1"
                affiche_err ("CAR_RPL_ISO-8859-1");
            else
            {
                // "certains caractres seront remplacs par des blancs\n"
                // "d'autres par ceux du jeu ISO-8859-1"
                affiche_err ("CAR_RPL_BLANCS");
                affiche_err ("CAR_RPL_ISO-8859b");
            }
        }

        // fin de la fonction dans ce cas
        return;
    }

    // charger en mmoire le fichier fonte
    chargefonte (fonte, 0x20);

    // on peut librer le fichier fonte
    fclose (ficfonte);
}



/* Charge une fonte  la place de iso-8859-1 si ncessaire */

void choix_fontebase ()
{
    char *sysfont;       // chemin d'accs au fichier sysfont
    char nomfonte [20];  // nom du fichier fonte mmoris dans sysfont
    FILE *descsysfont;   // descripteur du fichier sysfont
    int  caract;         // caractre du fichier sysfont
    int  i;              // compteur


    // recherche du fichier sysfont

    // si les commandes de cyloop sont implantes dans /usr/bin
    if (strcmp (dircom (), "/usr/bin") == 0)
    {
        // le fichier sysfont est dans /usr/share/cyloop
        sysfont = malloc (26);
        strcpy (sysfont, "/usr/share/cyloop/sysfont");
    }
    // sinon, le fichier sysfont est dans le
    // rpertoire bin des commandes excutables
    else
    {
        // calcul de la taille de mmoire ncessaire et allocation
        sysfont = malloc (strlen (dircom ()) + 9);

        // copie du rpertoire d'implantation des commandes de cyloop
        strcpy (sysfont, dircom ());

        // on rajoute /sysfont  la chaine prcdente
        strcpy (sysfont + strlen (sysfont), "/sysfont");
    }

    // si le fichier sysfont existe
    if (access (sysfont, 0) == 0)
    {
        // l'ouvrir en lecture
        descsysfont = fopen (sysfont, "r");

        // si ce fichier est accessible en lecture
        if (descsysfont)
        {
            i = 0;

            // lire le premier caractre de sysfont
            caract = getc (descsysfont);

            // tant que le dernier caractre lu
            // peut faire partie d'un nom de fichier
            while (caract > 0x20 && i < sizeof (nomfonte) - 1)
            {
                // mmoriser ce caractre aprs l'avoir mis en minuscules
                nomfonte [i++] = caract | 0x20;

                // lire le caractre suivant
                caract = getc (descsysfont);
            }

            // terminer la chaine de caractre contenue dans nomfonte
            nomfonte [i] = '\0';

            // on a fini d'utiliser le fichier sysfont
            fclose (descsysfont);

            // on peut charger la bonne fonte
            chargefonte_iso (nomfonte);
        }
        else
        {
            // "fichier sysfont prsent mais protg en lecture"
            // "le jeu de caractres ISO-8859-1 sera utilis"
            affiche_err ("SYSFONT_PROTEGE");
            affiche_err ("UTIL_ISO-8859-1");
        }
    }

    // on n'a besoin d'excuter cette fonction qu'une fois
    fonte_base_OK = 1;
}



/*
   insertion du dessin d'un caractre dans l'image

   paramtres : - adresse du dessin du caractre en mmoire
                - coordonnes du caractre dans l'image
*/

int dessinecar (octet *dessincar, int y, int x)
{
    octet pixligne;   // image du caractre sur la ligne en cours de traitement
    octet masque;     // comme son nom l'indique, masque sur les bits
    octet *posligne;  // adresse d'un pixel dans l'image
    int   i, j;       // compteurs


    // calculer l'adresse du coin haut gauche du caractre dans l'image
    posligne = image + (y * larg_totale) + x;

    // pour toutes les lignes du dessin du caractre
    for (i = 0; i < hauteur_fonte; i++)
    {
        // initialisation masque
        masque = 0x80;

        // extraire une ligne de pixels du dessin du caractre
        pixligne = dessincar [i];

        // pour tous les pixels de la ligne extraite
        for (j = 0; j < largeur_fonte; j++)
        {
            // si ce pixel est sombre
            if (pixligne & masque)
                // colorier ce pixel dans l'image
                *(posligne + j) = coulbase_texte;

            // passage au pixel suivant
            masque = masque >> 1;
        }

        // descendre d'une ligne dans l'image
        posligne = posligne - larg_totale;
    }
}



/*
   Fabrique une chaine de caractre dcrivant une squence UTF-8 incomplte

   Le contenu est exprim d'abord sous forme hexadcimale
   puis sous formes de caractes encods UTF-8
*/

char * affseq (octet carbase, octet * suiteseq, int resteseq)
{
    static char resultat [40]; // chaine de caractres gnre
    int  pospremier;           // position du premier caractre dans suiteseq
    int  numcar;               // compteur


    // premier caractre de la squence sous forme hexadcimale
    sprintf (resultat, "%02X", carbase);

    // si squence d'au moins 3 caractres
    // (la squence incomplte en comporte au moins 2)
    if (carbase >= 0xE0)
    {
        // position dans le tableau suiteseq du 2me caractre de la squence
        // (l'ordre des caractres est invers dans suiteseq)
        if (carbase < 0xF0)
            pospremier = 1;
        else
            pospremier = 2;

        // suite de la squence sous forme hexadcimale
        for (numcar = pospremier; numcar >= resteseq; numcar --)
            sprintf (resultat + strlen (resultat), " %02X", suiteseq [numcar]);
    }

    // premier caractre de la squence sous forme de caractre encod UTF-8
    sprintf (resultat + strlen (resultat), " (%c", carbase);

    if (carbase >= 0xE0)
        // suite de la squence sous forme de caractre encod UTF-8
        for (numcar = pospremier; numcar >= resteseq; numcar --)
            sprintf (resultat + strlen (resultat), "%%c", suiteseq [numcar]);

    // fin de la chaine gnre
    strcat (resultat, ")");

    // renvoyer le rsultat  la fonction appelante
    return (resultat);
}



/*
   insre dans l'image le caractre pass en paramtre aux
   coordonnes indiques

   en cas de besoin, charge en mmoire des jeux de caractres

   cette fonction traite les caractres cods UTF-8 sur plusieurs octets
*/

int affcar (octet caract, int y, int x)
{
    // mmorisation d'une squence UTF-8 ou fur et  mesure des appels de affcar
    static int   resteseq = 0; // caractres d'une squence UTF-8 restants
    static octet carbase;      // premier caractre d'une squence UTF-8
    static octet suiteseq [3]; // autres caractres dans une squence UTF-8

    // adresse de la reprsentation graphique d'un groupe de 64 caractres UTF-8
    octet  * sous_table;
    // reprsentation graphique d'un caractres UTF-8
    octet  dessincar [hauteur_fonte];


    // si caractre ASCII 7 bits ou on utilise un
    // jeu de caractres sur 8 bits dj en mmoire
    if (caract < 0x80 || fonte_base_OK)
    {
        // message d'erreur si une squence utf8 tait en cours
        if (resteseq)
        {
            // "Squence UTF-8 code %s incomplte"
       aff_err_arg ("SEQ_UTF8_TRONQUEE", affseq (carbase, suiteseq, resteseq));

            // on oublie cette squence
            resteseq = 0;
        }

        // si le caractre n'est pas un caractre de controle
        if (caract >= ' ')
            // insrer le dessin du caractre dans l'image
            dessinecar (fonte [caract - 0x20], y, x);

        // code de retour : un caractre a t rajout
        return 1;
    }
    // sinon si jeu de caractre UTF-8
    else if (util_utf8 ())
    {
        // si caractre entre C0 et FF
        // (premier caractre d'une squence UTF-8)
        if (caract >= 0xC0)
        {
            // message d'erreur si une autre squence utf8 tait dj en cours
            if (resteseq)
                // "Squence UTF-8 code %s incomplte"
       aff_err_arg ("SEQ_UTF8_TRONQUEE", affseq (carbase, suiteseq, resteseq));

            // mmoriser le premier caractre de la squence
            carbase = caract;

            // si on n'a pas encore charg la table ou ouvert le fichier
            // fonte associ aux squences commenant par ce caractre
            if (! init_utf [caract & 0x3F])
            {
                // le faire
                chargefonte_utf (carbase);

                // et mmoriser l'opration (qu'elle ait fonctionn ou non)
                init_utf [caract & 0x3F] = 1;
            }

            // calculer le nombre de caractres restant dans la squence
            if (caract < 0xE0)
                resteseq = 1;
            else if (caract < 0xF0)
                resteseq = 2;
            else
                resteseq = 3;

            // code de retour : aucun caractre n'a t rajout dans l'image
            return 0;
        }
        // sinon caractre entre 80 et BF
        // (2me, 3me ou 4me caractre d'une squence UTF-8)
        else
        {
            // si on est dans une squence de caractres UTF-8
            if (resteseq)
            {
                // mmoriser le caractre
                suiteseq [--resteseq] = caract;

                // si toute la squence a t mmorise
                if (! resteseq)
                {
                    // si squence UTF-8 sur 2 caractres (fonte en mmoire)
                    if (carbase < 0xE0)
                    {
                        // slectionner la table en mmoire qui
                        // contient le dessin du caractre
                        sous_table = table_utf [carbase & 0x1F];

                        // insrer le dessin du caractre dans l'image
                        dessinecar (sous_table + ((*suiteseq & 0x3F)
                                                    * hauteur_fonte), y, x);
                    }
                    // sinon squence UTF-8 sur 3 ou 4 caractres
                    else
                    {
                        // rcuprer la reprsentation graphique
                        // du caractre dans le fichier fonte
                        recupfonte (carbase, suiteseq, dessincar);

                        // insrer le dessin du caractre dans l'image
                        dessinecar (dessincar, y, x);
                    }

                    // code de retour : un caractre a t rajout
                    return 1;
                }
                else
                    // code de retour : aucun caractre rajout dans l'image
                    return 0;
            }
            // sinon
            else
            {
                // caractre %02X (%c) en dehors d'une squence UTF-8
                printf (message ("CAR_HORS_SEQ_UTF8"), caract, caract);
                putchar ('\n');

                // code de retour : aucun caractre rajout dans l'image
                return 0;
            }
        }
    }
    // sinon (jeu de caractres sur 8 bits qui n'est pas encore en mmoire)
    else
    {
        // charger en mmoire le jeu de caractres
        // (pour la partie qui diffre du jeu ISO-8859-1)
        choix_fontebase ();

        // insrer le dessin du caractre dans l'image
        dessinecar (fonte [caract - 0x20], y, x);

        // code de retour : un caractre a t rajout
        return 1;
    }
}



/*
   insre dans l'image la chaine de caractres passe
   en paramtre  partir des coordonnes indiques

   vrifie auparavant les coordonnes demandes pour la chaine
*/

void aff_texte (octet *chaine, int y, int x, int alignement)
{
    int   longueur;  // longueur de la chaine de caractres
    octet *c;        // pointeur sur un caractre de la chaine


    // si criture demand trop bas ou trop haut dans l'image
    if (y < hauteur_fonte || y > haut_totale)
        // on n'crira rien
        return;

    // si alignement  droite ou centr
    if (alignement != align_gauche)
    {
        // on commence par calculer la longueur de la chaine

        // si jeu de caractre UTF-8
        if (util_utf8 ())
        {
            // initialisation
            c = chaine;
            longueur = 0;

            // dterminer la longueur de la chaine
            while (*c)
            {
                // compter les caractres ASCII et le premier
                // octet des codes UTF-8 sur plusieurs octets
                if (*c <= 0x7F || *c >= 0xC0)
                    longueur++;

                // avancer d'un caractre dans la chaine
                c++;
            }
        }
        // sinon
        else
            // la longueur de la chaine est plus facile  obtenir
            longueur = strlen (chaine);

        // changement d'unit, la longueur de la chaine est compte en pixels
        longueur = largeur_fonte * longueur;

        // calculer les coordonnes du caractre de la chaine le plus  gauche
        if (alignement == align_droite)
            x = x - longueur;
        else // alignement centr
            x = x - (longueur / 2);
    }

    // si le dbut de la chaine est trop  gauche dans l'image
    if (x < 0)
    {
        // on va sauter les premiers caractres pour essayer d'afficher le reste

        // si jeu de caractre UTF-8
        if (util_utf8 ())
        {
            // parcourir la chaine
            while (*chaine && x < 0)
            {
                // en comptant les caractres  afficher
                if (*chaine <= 0x7F || *chaine >= 0xC0)
                    x = x + largeur_fonte;

                chaine ++;
            }

            // sauter la fin des codes UTF-8 sur plusieurs octets
            while (*chaine >= 0x80 && *chaine <= 0xBF)
                chaine ++;
        }
        // sinon (caractres ASCII ou ISO-8859 ou similaire)
        else
        {
            // progression dans la chaine plus simple
            while (*chaine && x < 0)
            {
                x = x + largeur_fonte;
                chaine ++;
            }
        }
    }

    // tant que chaine non recopie et il reste de la place dans la ligne
    while (*chaine && (x + largeur_fonte) < larg_totale)
    {
        // afficher un caractre
        if (affcar (*chaine, y, x))
            // et comptabiliser le dplacement
            // dans le cas du codage UTF-8, plusieurs octets peuvent
            // tre ncessaire pour reprsenter un caractre
            x = x + largeur_fonte;

        // passer au caractre suivant dans la chaine  recopier
        chaine ++;
    }
}



/*
   insre dans l'image la valeur numrique entire passe
   en paramtre  partir des coordonnes indiques
*/

void aff_entier (long valeur, int y, int x, int alignement)
{
    char chaine [12];  // chaine contenant la valeur numrique


    // convertir la valeur numrique en chaine de caractres
    sprintf (chaine, "%d", valeur);

    // afficher cette chaine de caractres
    aff_texte (chaine, y, x, alignement);
}



/*
   insre dans l'image la valeur numrique flottante passe
   en paramtre  partir des coordonnes indiques
*/

void aff_float (float valeur, char *format, int y, int x, int alignement)
{
    char conversion [10]; // format de conversion
    char chaine [20];     // chaine contenant la valeur numrique convertie


    // gnrer le paramtre format complet pour convertir la valeur numrique
    if (*format == '%')
        strcpy (conversion, format);
    else
    {
        *conversion = '%';
        strcpy (conversion + 1, format);
    }

    // convertir la valeur numrique en chaine de caractres
    sprintf (chaine, conversion, valeur);

    // afficher cette chaine de caractres
    aff_texte (chaine, y, x, alignement);
}
